///////////////////////////////////////////////////////////
//
// simple OpenCV and GLSL demo program for CS 525 GPU Programming
// 
// Written by Andy Johnson 12/11
// originally based off the Lighthouse 3D tutorials
///////////////////////////////////////////////////////////

#include "GL/glew.h"

#ifdef __linux__
#include <GL/glut.h>
#endif

#ifdef __APPLE__
#include <glut.h>
#endif

#ifdef _WIN32
#include "glut.h"
#endif

#include "cv.h"
#include "highgui.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define IMAGE_SIZE 800

int winWidth = IMAGE_SIZE;
int winHeight = IMAGE_SIZE;

GLuint chosenTex, frame;

// openCV stuff
IplImage *image = 0;
IplImage *destination = 0;
IplImage * showMe;

GLuint cameraImageTextureID;
CvCapture *capture = 0;


////////////////////////////////////////////////////////////////////////////////

char *textFileRead(char *fn)
{
	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
			fseek(fp, 0, SEEK_END);
			count = ftell(fp);
			rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	
	if (content == NULL)
	   {
	   fprintf(stderr, "ERROR: could not load in file %s\n", fn);
	   exit(1);
	   }
	return content;
}                    


////////////////////////////////////////////////////////////////////////////////

GLuint p, p2;

float lpos[4] = {1.0, 0.5, 1.0, 0.0};

void changeSize(int w, int h) {

	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	
	if(h == 0)
		h = 1;

	float ratio = 1.0* winWidth / winHeight;

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	winWidth = w; //winWidth;
	winHeight = h; // winWidth;
	
	// Set the viewport to be the entire window
    glViewport(0, 0, winWidth, winHeight);

	// Set the correct perspective.
	gluPerspective(45, ratio, 1, 1000);
	glMatrixMode(GL_MODELVIEW);
}
	
////////////////////////////////////////////////////////////////////////////////

void printShaderLog(GLuint prog)
{
    GLint infoLogLength = 0;
    GLsizei charsWritten  = 0;
    GLchar *infoLog;

    glGetShaderiv(prog, GL_INFO_LOG_LENGTH, &infoLogLength);

    if (infoLogLength > 0)
    {
        infoLog = (char *) malloc(infoLogLength);
        glGetShaderInfoLog(prog, infoLogLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

void printProgramLog(GLuint shad)
{
    GLint infoLogLength = 0;
    GLsizei charsWritten  = 0;
    GLchar *infoLog;

    glGetProgramiv(shad, GL_INFO_LOG_LENGTH, &infoLogLength);

    if (infoLogLength > 0)
    {
        infoLog = (char *) malloc(infoLogLength);
        glGetProgramInfoLog(shad, infoLogLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}


////////////////////////////////////////////////////////////////////////////////


void renderScene(void) {
    GLint texLoc, effectLoc;
    int xgrid, ygrid, counter;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();
	gluLookAt(0.0,0.0,5.0, 
		      0.0,0.0,-1.0,
			  0.0f,1.0f,0.0f);
	glColor3f(1.0, 1.0, 1.0);	
	
    // openCV stuff
    IplImage *frame = 0;

    frame = cvQueryFrame( capture );

	// If I can't get a frame then exit the program
	if(!frame) 
		{
		exit(1);
		}

	// create a guaranteed texture size
	destination = cvCreateImage ( cvSize(800, 600), frame->depth, frame->nChannels );
	cvResize(frame, destination);

	// set the texture parameters
	glBindTexture( GL_TEXTURE_2D, cameraImageTextureID); //bind the texture to it's array
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

	showMe = destination;

	if(showMe->nChannels == 3)
		{
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB, showMe->width, showMe->height, 0, GL_BGR, GL_UNSIGNED_BYTE, showMe->imageData);
		} 
	else if(frame->nChannels == 4)
		{
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA, showMe->width, showMe->height, 0, GL_BGRA, GL_UNSIGNED_BYTE, showMe->imageData);
		}

	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, cameraImageTextureID);
	
	// use the shader
	glUseProgram(p);
	
	// always using texture 0 for now
    texLoc = glGetUniformLocation(p, "whichTexture");
    glUniform1i(texLoc, 0);


	// by default draw a white polygon if the texturing doesn't work.
    glColor3f(1.0, 1.0, 1.0);
    counter = 0;
    
    // draw the textured polygons
    // for some more sophisitcated effects you will probably want to render the texture to an offscreen
    // buffer first to be able to chain multiple shader effects together
        
    for (xgrid=0; xgrid < 2; xgrid++)
    	for (ygrid=0; ygrid < 2; ygrid++)
			{
				// pick the effect to use
				effectLoc = glGetUniformLocation(p, "whichEffect");
				glUniform1i(effectLoc, counter);
				counter++;

			// draw the polygon with the current texture and effect
		
			glBegin( GL_POLYGON );    
				// GL_TEXTURE_RECTANGLE_ARB uses pixel coordinates rather than the standard OpenGL 0.0 to 1.0
				//glTexCoord2f( 0.0f, 1.0f );
				glTexCoord2i(0, showMe->height);
				glVertex3f( -2.0f+2*xgrid, -2.0f+2*ygrid, 0.0f );
				
				//glTexCoord2f( 1.0f, 1.0f ); 
				glTexCoord2i(showMe->width, showMe->height);
				glVertex3f(  0.0f+2*xgrid, -2.0f+2*ygrid, 0.0f );
				
				//glTexCoord2f( 1.0f, 0.0f );
				glTexCoord2i(showMe->width, 0);
				glVertex3f(  0.0f+2*xgrid,  0.0f+2*ygrid, 0.0f );
				
				//glTexCoord2f( 0.0f, 0.0f ); 
				glTexCoord2i(0, 0);
				glVertex3f( -2.0f+2*xgrid,  0.0f+2*ygrid, 0.0f );
		   glEnd();
		}
	glutSwapBuffers();
}

////////////////////////////////////////////////////////////////////////////////

void processNormalKeys(unsigned char key, int x, int y) {

	// exit program
	if (key == 27) 
		exit(0);	
}

////////////////////////////////////////////////////////////////////////////////

GLuint setShaders(char * vert, char * frag) {
    GLuint v,f, pro;
	char *vs,*fs;

	v = glCreateShader(GL_VERTEX_SHADER);
	f = glCreateShader(GL_FRAGMENT_SHADER);

	vs = textFileRead(vert);
	fs = textFileRead(frag);

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(v, 1, &vv,NULL);
	glShaderSource(f, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShader(v);
	glCompileShader(f);

	printShaderLog(v);
	printShaderLog(f);

	pro = glCreateProgram();
	glAttachShader(pro,v);
	glAttachShader(pro,f);

	glLinkProgram(pro);
	printProgramLog(pro);
	
	return(pro);
}


////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv) {

	// initialize camera or reading from video file
	capture = cvCaptureFromCAM( 0 );
	//capture = cvCaptureFromFile("some file.m4v");  // could also capture from a file


	if( !capture ) 
	{
	printf( "Cannot get access to the video feed (from file or camera)!\n" );
	return 1;
	}

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(winWidth, winHeight);
	glutCreateWindow("GLSL");

	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutKeyboardFunc(processNormalKeys);

	glewInit();
	
	glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_RECTANGLE_ARB);

	glClearColor(0.0,0.0,0.0,1.0);

 

	if (GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader)
		printf("Ready for GLSL\n");
	else {
		printf("No GLSL support\n");
		exit(1);
	}

	p = setShaders((char *) "./webcam.vert", (char *) "./webcam.frag");
	
	glGenTextures(1, &cameraImageTextureID);
	glActiveTexture(GL_TEXTURE0 + 0); //GL_TEXTURE0
	
	glutMainLoop();
	return 0;
}
